#!/usr/bin/env python3

import time
import struct
import sys

out = False
if len(sys.argv) == 1:
	print(file=sys.stderr)
	print('Usage: python3 extract_h264.py fileIn [... more files]', file=sys.stderr)
	print(file=sys.stderr)
	sys.exit(1)

for i in range(1, len(sys.argv)):
	fileIn = sys.argv[1]

	fIn = open(fileIn,'rb')

	while True:
		try:
			zero, time, length = struct.unpack('!BQQ', fIn.read(17))
			data = fIn.read(length)
			sys.stdout.buffer.write(b'\0\0\0\1' + data)
		except:
			break;
	
	fIn.close()




#!/usr/bin/env python3

import logging
import time
import glob
import struct
import sys
import os
import re
import live555
import threading
from datetime import datetime
from shared import common

if common.check_prereqs() == False:
	print("Prerequisite check failed: {}".format(common.get_prereqs_msg()))
	exit(-1)

if len(sys.argv) != 7:
	print("ERROR : usage: rtsp.py chunk_path jpg_path url deviceid initial_timeout frame_timeout")
	exit(-1)

chunk_path = sys.argv[1]
jpg_path = sys.argv[2]
url = sys.argv[3]
deviceid = sys.argv[4]
initial_timeout = int(sys.argv[5])
frame_timeout = int(sys.argv[6])

logger = logging.getLogger("rtsp")
logging.basicConfig(filename="/tmp/{}.log".format(deviceid),level=logging.DEBUG)

basepath = os.path.dirname(os.path.abspath(__file__))

fixedurl = re.sub(r'rtsp:\/\/.*@', 'rtsp://........@', url, flags=re.IGNORECASE)
logger.info("Starting recording from {} to {} for {}".format(fixedurl, chunk_path,deviceid))

class createFrameThread(threading.Thread):
	def __init__(self, video, jpg, thumb, mp4):
		threading.Thread.__init__(self)
		self.video = video
		self.jpg = jpg
		self.thumb = thumb
		self.mp4 = mp4
	def run(self):
		global logger
		try:
			p2 = common.subprocess_chunk_to_mp4(basepath, self.video, self.mp4, "")
			p = common.subprocess_chunk_to_jpg(basepath, self.video, self.jpg, "")
			p.wait()
			p = common.subprocess_resize_jpg(self.jpg, self.thumb, "scale=320:-1")
			p.wait()
			p2.wait()
			os.remove(self.video)

			if os.path.isfile(self.mp4) and os.stat(self.mp4).st_size < 4096:
				os.remove(self.mp4)
		except BaseException as e: 
			logger.error("CREATE FRAME THREAD - ", e)
		except:
			logger.error("CREATE FRAME THREAD - Something bad happend")


def createFrame(video, jpg, thumb, mp4):
	if video is None or len(video) == 0:
		return

	t = createFrameThread(video, jpg, thumb, mp4)
	t.start()
	return t

def record(chunk_path, jpg_path, url, deviceid, initial_timeout, frame_timeout):
	global logger

	now = datetime.now()
	chunk_filename = common.getChunkFilenameBy(chunk_path, deviceid, now)
	jpg_filename = common.getJPGFilenameBy(jpg_path, deviceid, now)
	thumb_filename = common.getThumbnailFilenameBy(jpg_path, deviceid, now)
	mp4_filename = common.getMP4FilenameBy(chunk_path, deviceid, now)

	state = {
		'lastfilename': "",
		'lastjpgfilename': "",
		'lastthumbfilename': "",
		'lastmp4filename': "",
		'fOut': open(chunk_filename, 'wb'),
		'filename': chunk_filename,
		'jpgfilename': jpg_filename,
		'thumbfilename': thumb_filename,
		'mp4filename': mp4_filename,
		'last_update': now,
		'first_update': True
	}

	# Cleanup and convert old chunk files for this camera into mp4's if done
	oldchunkfiles = glob.glob(os.path.join(chunk_path,"*/{}.*.*.chunk".format(deviceid)))
	if chunk_filename in oldchunkfiles: oldchunkfiles.remove(chunk_filename)
	for oldchunkfile in oldchunkfiles:
		logger.info("Upgrading {}".format(oldchunkfile))
		date_object = datetime.strptime(os.path.basename(oldchunkfile), "{}.%Y%m%d.%H%M.chunk".format(deviceid))
		old_jpg_filename = common.getJPGFilenameBy(jpg_path, deviceid, date_object)
		old_thumb_filename = common.getThumbnailFilenameBy(jpg_path, deviceid, date_object)
		old_mp4_filename = common.getMP4FilenameBy(chunk_path, deviceid, date_object)
		createFrame(oldchunkfile, old_jpg_filename, old_thumb_filename, old_mp4_filename).join()

	

	def oneFrame(codecName, bytes, sec, usec, durUSec, ctx):
		ctx['last_update'] = datetime.now()
		delta = datetime.now() - datetime(1970, 1, 1)
		frame_time = int(delta.total_seconds()*1000000)
		ctx['first_update'] = False

		F = int(bytes[0] & 128)
		NRI = int((bytes[0] & (64+32) ) / 32)
		TYPE = int((bytes[0] & (16+8+4+2+1) ))

		if TYPE==7:
			now = datetime.now()
			chunk_filename = common.getChunkFilenameBy(chunk_path, deviceid, now)
			jpg_filename = common.getJPGFilenameBy(jpg_path, deviceid, now)
			thumb_filename = common.getThumbnailFilenameBy(jpg_path, deviceid, now)
			mp4_filename = common.getMP4FilenameBy(chunk_path, deviceid, now)
			ctx['filename'] = chunk_filename
			ctx['jpgfilename'] = jpg_filename
			ctx['thumbfilename'] = thumb_filename
			ctx['mp4filename'] = mp4_filename

		if ctx['filename'] != ctx['lastfilename']:
			ctx['fOut'].close()
			createFrame(ctx['lastfilename'], ctx['lastjpgfilename'], ctx['lastthumbfilename'], ctx['lastmp4filename'])
			logger.info('INFO : opening {}'.format(ctx['filename']))
			ctx['fOut'] = open(ctx['filename'], 'wb')
			ctx['lastfilename'] = ctx['filename']
			ctx['lastjpgfilename'] = ctx['jpgfilename']
			ctx['lastthumbfilename'] = ctx['thumbfilename']
			ctx['lastmp4filename'] = ctx['mp4filename']


		ctx['fOut'].write(struct.pack('!BQQ', 0, frame_time, len(bytes)))
		ctx['fOut'].write(bytes)
		ctx['fOut'].flush()

		return ctx

	temp_oneFrame = lambda codecName, bytes, sec, usec, durUSec: oneFrame(codecName, bytes, sec, usec, durUSec, state)

	# Starts pulling frames from the URL, with the provided callback:
	useTCP = False
	live555.startRTSP(url, temp_oneFrame, useTCP)

	# Run Live555's event loop in a background thread:
	t = threading.Thread(target=live555.runEventLoop, args=())
	t.setDaemon(True)
	t.start()

	while True:
		delta = datetime.now() - state['last_update']
		time_since_update = delta.total_seconds()
		if state['first_update'] and time_since_update >= initial_timeout:
			logger.warn("Took longer than {} seconds for first frame".format(initial_timeout))
			break; 
		if state['first_update'] == False and time_since_update >= frame_timeout:
			logger.warn("Delay of over {} seconds for data".format(frame_timeout))
			break; 
		time.sleep(1)

	# Tell Live555's event loop to stop:
	live555.stopEventLoop()

	# Wait for the background thread to finish:
	t.join()

record(chunk_path, jpg_path, url, deviceid, initial_timeout, frame_timeout)




#!/usr/bin/env python3

import logging
from shared import nvrconfig
from shared import reap
from shared import common
from shared.camera import Camera, minutesBetween, mergeRanges
from pprint import pprint
import threading
from datetime import datetime,timedelta
from subprocess import call,check_call
from bottle import route, run, template, static_file, HTTPError, request, response, install
from os import listdir
from os.path import isfile, join, exists, isdir
import glob
import os
import signal
import time
import subprocess
import sys
import atexit
import json
import static
from shared import daemon



if common.check_prereqs() == False:
	print("Prerequisite check failed: {}".format(common.get_prereqs_msg()))
	exit(-1)


LOG_FILE="/tmp/pynvr.log"
PID_FILE="/tmp/pynvr.pid"


basepath = os.path.dirname(os.path.abspath(__file__))
static_dir=static.get_static_root()

#if 'PYTHONPATH' not in os.environ.keys() or len(os.environ['PYTHONPATH']) == 0:
	#print("ERROR : PYTHONPATH is not set, recording will not work without it pointing to pylive555/build/lib.linux-x86_64-3.4 or something similar.", file=sys.stderr)
	#exit(-1)

class MyDaemon(daemon.daemon):
	def __init__(self, pidfile, logger):
		daemon.daemon.__init__(self, pidfile)
		self.logger = logger
	def run(self):
		global reaper,basepath, static_dir

		while True:

			self.logger.info("PyNVR - Python Network Video Recorder")
			self.logger.info("-------------------------------------")

			self.logger.info("Basepath: {}".format(basepath))
			self.logger.info("Static: {} ".format(static_dir))

			config_file = nvrconfig.locate_config_file('nvr.json')
			self.logger.info("config: {}".format(config_file))
			config = nvrconfig.parse(config_file, self.logger)

			if config == None:
				self.logger.error("There was a problem parsing the config file")
				exit(-1);

			common.setConfig(config)

			reaper = None

			class reaperThread(threading.Thread):
				def __init__(self, config, logger):
					threading.Thread.__init__(self)
					self.config = config
					self.done = False
					self.logger = logger
				def run(self):
					self.logger.info("Reaper starting . . .")
					while self.done == False:
						while reap.reap(self.config, logger) and self.done == False:
							True

						for a in range(0,60):
							time.sleep(.25)
							if self.done:
								break;
					self.logger.info("Reaper stopped.")
				def finish(self):
					self.done = True


			# This thread records a stream of data from a camera
			# If finish is called it kills it's spawned process and exits
			class recordThread (threading.Thread):
				def __init__(self, config, logger):
					threading.Thread.__init__(self)
					self.config = config
					self.done = False
					self.logger = logger
				def run(self):
					if 'initial_timeout' in self.config.keys():
						initial_timeout = self.config['initial_timeout']
					else:
						initial_timeout = 30
					if 'frame' in self.config.keys():
						frame_timeout = self.config['frame_timeout']
					else:
						frame_timeout = 5
					cmd = '{}/rtsp.py "{}" "{}" "{}" "{}" "{}" "{}" '.format(basepath, self.config['dst'],config['system']['jpg_cache_dir'],self.config['url'],self.config['id'], initial_timeout, frame_timeout)
					while self.done == False:
						self.logger.info("Starting record process for {}".format(self.config['id']))
						self.process = subprocess.Popen(cmd, shell=True, preexec_fn=os.setsid)
						while self.process.poll() is None and self.done == False:
							time.sleep(0.25)

						if (self.process.returncode != 0 and self.done == False):
							self.logger.error("There was a problem spawning the recording process for camera {}, return code is {}, trying again in a minute.".format(self.config['id'], self.process.returncode))
							for a in range(0,240):
								time.sleep(.25)
								if self.done:
									break;
				def finish(self):
					self.done = True
					try:
						os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)
					except ProcessLookupError as e:
						self.logger.warn("Process not found: errno = {}, msg = {}".format(e.errno, e.strerror))	

			# Catch ctrl-c and shutdown nicely
			def signal_handler(signal, frame):
				global w
				global config
				self.logger.warn('You pressed Ctrl+C!')
				sys.exit(0)

			def exit_handler():
				global config
				self.logger.info('Shutting down')
				stopThreads(common.getConfig())

			atexit.register(exit_handler)
			signal.signal(signal.SIGINT, signal_handler)
			signal.signal(signal.SIGTERM, signal_handler)

			# Start all threads
			def startThreads(config):
				global reaper
				self.logger.info("Starting Recording Threads . . .")
				for camera in config['cameras']:
					if camera['enabled']:
						self.logger.info("starting camera: id = {}".format(camera['id']))
						thread = recordThread(camera, self.logger)
						thread.daemon = True
						thread.start()
						camera['_thread'] = thread

				if reaper == None:
					if 'reaper' in common.getConfig().keys():
						reaper = reaperThread(common.getConfig()['reaper'], self.logger)
						reaper.daemon = True
						reaper.start()

			def stopThreads(config):
				global reaper
				self.logger.info("Stopping Recording Threads . . .")
				for camera in config['cameras']:
					if camera['enabled']:
						if '_thread' in camera.keys():
							self.logger.info("stopping camera: id = {}".format(camera['id']))
							camera['_thread'].finish()
							camera.pop('_thread', None)
				if reaper != None:
					reaper.finish()
					reaper = None

				time.sleep(1)

			class  watcherThread(threading.Thread):
				def __init__(self):
					threading.Thread.__init__(self)
				def run(self):
					startThreads(common.getConfig())

					while [ True ]:
						start_mtime = os.stat(config_file).st_mtime
						while True:
							try:
								if start_mtime != os.stat(config_file).st_mtime:
									self.logger.info("config file change. reloading . . .\n")
									temp_config = nvrconfig.parse(config_file, self.logger)
									if (temp_config == None):
										self.logger.error("Config file changed, but could not be loaded, continuing with older config.")
										time.sleep(1)
									else:
										stopThreads(common.getConfig())
										config = temp_config
										common.setConfig(config)
										time.sleep(1)
										startThreads(config)
									break
							except FileNotFoundError as e:
								time.sleep(1)

			def payload(r):
				return json.dumps({ 'payload' : r, 'date': datetime.now().strftime("%Y-%m-%dT%H:%M:%S"), 'system' : { 'name' : 'PyNVR', 'version' : '0.1' } })

			@route('/api/1/system.json')
			def system():
				running = True
				camera_status = {}
				for cam in config['cameras']:
					if cam['enabled']:
						value = Camera(common.getCameraConfig(cam['id'])).isRecording()
						camera_status[cam['id']] = value
						if not value:
							running = False


				r = { "status" : "running" if running else "error", "cameras" : camera_status }
				return payload(r)

			@route('/api/1/cameras.json')
			def cameras():
				r = []
				for cam in config['cameras']:
					if cam['enabled']:
						r.append(cam['id'])
				return payload(r)

			def fmt(d):
				return d.strftime("%Y%m%d.%H%M")


			@route('/api/1/camera/<camera_id>.json')
			def camera(camera_id):
				c = common.getCameraConfig(camera_id)

				video = c['dst']
				mp4s = [f for f in listdir(video) if isfile(join(video, f)) and f.endswith('mp4')]

				dates = [datetime.strptime(c, "{}.%Y%m%d.%H%M.mp4".format(camera_id)) for c in mp4s]
				dates.sort()

				ranges = []

				if len(dates) > 0:
					begin=min(dates)
					lastdate=begin
					for date in dates:
						diff = abs(lastdate - date)
						minutes=(int(diff.total_seconds()/60))
						if (minutes>1):
							ranges.append({ 'from' : fmt(begin), 'to' : fmt(lastdate) })
							begin=date
						lastdate = date

					if (begin != lastdate):
						ranges.append({ 'from' : fmt(begin), 'to' : fmt(lastdate) })

				r = { 
					"camera" : {
						"id" : camera_id,
						"enabled" : c['enabled'],
						"dates" : ranges
					}
				}
				return payload(r)

			@route('/api/1/camera/<camera_id>.preview.<YYYYMMDD>.<HHMM>.jpg')
			def camera_preview(camera_id, YYYYMMDD, HHMM):
				global config
				c = common.getCameraConfig(camera_id)
				jpg_filename = common.getJPGFilename(common.getConfig(), c, YYYYMMDD + "." + HHMM)

				return static_file(os.path.basename(jpg_filename), root = os.path.dirname(jpg_filename))

			@route('/api/1/camera/<camera_id>.thumbnail.<YYYYMMDD>.<HHMM>.jpg')
			def camera_thumbnail(camera_id, YYYYMMDD, HHMM):
				global config
				c = common.getCameraConfig(camera_id)
				jpg_filename = common.getJPGFilename(common.getConfig(), c, YYYYMMDD + "." + HHMM)
				if not os.path.isfile(jpg_filename):
					mp4_filename = common.getMP4Filename(c, YYYYMMDD + "." + HHMM)
					common.subprocess_mp4_to_jpg(mp4_filename, jpg_filename, None).wait()
				thumb_filename = common.getThumbnailFilename(common.getConfig(), c, YYYYMMDD + "." + HHMM)
				if not os.path.isfile(thumb_filename):
					common.subprocess_resize_jpg(jpg_filename, thumb_filename, "scale=320:-1").wait()

				return static_file(os.path.basename(thumb_filename), root = os.path.dirname(thumb_filename))

			@route('/api/1/camera/<camera_id>.<YYYYMMDD>.<HHMM>.mp4')
			def camera_mp4(camera_id, YYYYMMDD, HHMM):
				global config
				c = common.getCameraConfig(camera_id)
				mp4_filename = common.getMP4Filename(c, YYYYMMDD + "." + HHMM)

				return static_file(os.path.basename(mp4_filename), root = os.path.dirname(mp4_filename))

			@route('/api/1/camera/<camera_id>.range.<FROM_YYYYMMDD.HHMM>.to.<TO_YYYYMMDD.HHMM>.mp4')
			def camera_movie_range(camera_id, FROM_YYYYMMDD_HHMM, TO_YYYYMMDD_HHMM):
				c = common.getCameraConfig(camera_id)
				f = datetime.strptime(FROM_YYYYMMDD_HHMM, "%Y%m%d.%H%M")
				t = datetime.strptime(TO_YYYYMMDD_HHMM, "%Y%m%d.%H%M")
				mp4_filename = common.getMP4FilenameRange(common.getConfig(), c, FROM_YYYYMMDD_HHMM, FROM_YYYYMMDD_HHMM)

				files = []
				diff = int((t - f).total_seconds() / 60)
				for a in range(0,diff):
					d = f + timedelta(minutes = a)
					chunk_filename = common.getChunkFilename(c, d)
					if isfile(chunk_filename):
						files.append(chunk_filename)

					if len(files) > 60:
						raise HTTPError(413, "Time range too large, resulting movie would be too large")
						

				p = common.subprocess_chunks_to_mp4(basepath, files, mp4_filename)
				p.wait()

				return static_file(os.path.basename(mp4_filename), root = os.path.dirname(mp4_filename))

			@route('/api/1/cameras/<YYYYMMDD>')
			def camera_daily_info(YYYYMMDD):

				camera_ids = request.params['camera_ids']

				cameras = {}
				for camera_id in camera_ids.split(','):
					cameras[camera_id] = Camera(common.getCameraConfig(camera_id))

				dateRanges = []
				for camera_id in camera_ids.split(','):
					dateRanges.extend(cameras[camera_id].dateRangesForDay(datetime.strptime(YYYYMMDD,"%Y%m%d")))

				ranges = mergeRanges(dateRanges)

				if len(dateRanges) == 0:
					return payload({'dates' : [], 'min' : fmt(datetime.now()), 'max' : fmt(datetime.now()), 'camera_ids': camera_ids.split(',')})

				min_date = ranges[0].begin
				max_date = ranges[len(ranges)-1].end

				values = {} 
				for r in ranges:
					for d in minutesBetween(r):
						chunks = []
						for camera_id in camera_ids.split(','):
							chunk = cameras[camera_id].getChunk(d)
							if chunk != None:
								chunks.append(camera_id)
						values[fmt(d)] = chunks

				return payload({'dates' : values, 'min' : fmt(min_date), 'max' : fmt(max_date), 'camera_ids': camera_ids.split(',')})

			@route('/static/<path:path>')
			def static(path):
				return static_file(path, root = static_dir)

			@route('/')
			def index():
				return static_file("web.html", root = static_dir)

			w = watcherThread()
			w.daemon = True
			w.start()

			def plugin_logger(func):
				def wrapper(*args, **kwargs):
					global logger
					logger.info('%s %s %s %s %s' % (request.remote_addr, datetime.now().strftime('%H:%M'), request.method, request.url, response.status))
					req = func(*args, **kwargs)
					return req
				return wrapper

			install(plugin_logger)


			logger.info("starting bottle. . . ")
			run(host=config['system']['server']['host'], port=config['system']['server']['port'], quiet = True)

			time.sleep(10)


logger = logging.getLogger("pynvr")
#logger.setLevel(logging.INFO)
#formatter = logging.Formatter("%(ascitime)s - %(name)s - %(levelname)s - %(message)s")
#formatter = logging.Formatter("%(ascitime)s - %(name)s - %(levelname)s - %(message)s")
#handler = logging.FileHandler(LOG_FILE)
#handler.setFormatter(formatter)
#logger.addHandler(handler)

logging.basicConfig(filename=LOG_FILE,level=logging.DEBUG)



if __name__ == "__main__":
	daemon = MyDaemon('/tmp/pynvr.pid', logger)
	if len(sys.argv) == 2:
		if 'start' == sys.argv[1]:
			daemon.start()
		elif 'stop' == sys.argv[1]:
			daemon.stop()
		elif 'restart' == sys.argv[1]:
			daemon.restart()
		elif 'status' == sys.argv[1]:
			daemon.status()
		elif 'foreground' == sys.argv[1]:
			ch = logging.StreamHandler(sys.stdout)
			ch.setLevel(logging.DEBUG)
			formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
			ch.setFormatter(formatter)
			logging.getLogger().addHandler(ch)

			daemon.run()
		else:
			print("Unknown command", file=sys.stderr)
			sys.exit(2)
		sys.exit(0)
	else:
		print("usage: {} start|stop|restart|foreground".format(sys.argv[0]), file=sys.stderr)
		sys.exit(2)



#!/usr/bin/env python3

import os
import json

def _input(prompt, default):
	if len(default) == 0:
		return input(prompt)
	result = input('{} ({})'.format(prompt, default))
	return default if len(result) == 0 else result

def _reap_stanza(disk, extensions):
	return {
		"disk": disk,
		"percent" : "95",
		"extensions" : extensions
	}

print("PyNVR Configuration utility")
print("---------------------------")
print("This utility will ask a few questions and create a nvr.json file for you to use.")
print("")

base = input('Base Directory for storing files: ')

reaper = []

server_name = _input('Server name:', 'localhost')
server_port = int(_input('Server port:', '8080'))
jpg_cache_dir = _input('Directory for storing temporary jpg files:', os.path.join(base, "working", "jpg"))
mp4_cache_dir = _input('Directory for storing temporary jpg files:', os.path.join(base, "working", "mp4"))
camera_logs_dir = _input('Directory for storing temporary jpg files:', os.path.join(base, "log"))
log_dir = _input('Directory for storing temporary jpg files:', os.path.join(base, "log"))

num = int(_input('How many cameras to configure:','1'))

reaper.append(_reap_stanza(jpg_cache_dir,["jpg"]))
reaper.append(_reap_stanza(mp4_cache_dir,["mp4"]))


cameras = []
for a in range(0,num):
	num = a + 1
	name = _input("Camera {} name:".format(num), "camera{}".format(num))
	url = _input("Url for Camera {}:".format(name), "")
	dst = _input("Path to store video for camera {} url:".format(name), os.path.join(base, "cameras", name))
	initial_timeout = int(_input("Initial timeout for camera {}:".format(name), "30"))
	frame_timeout = int(_input("Frame timeout for camera {}:".format(name), "5"))
	cameras.append({
		"enabled": True,
		"id" : name,
		"url" : url,
		"dst" : dst,
		"initial_timeout" : initial_timeout,
		"frame_timeout" : frame_timeout
	})
	reaper.append(_reap_stanza(dst,["mp4"]))
	



data = {
	"system" : {
		"jpg_cache_dir" : jpg_cache_dir,
		"mp4_cache_dir" : mp4_cache_dir,
		"camera_log_dir" : camera_logs_dir,
		"log_dir" : log_dir,
		"server" : {
			"host" : server_name,
			"port" : server_port
		}
	},
	"cameras" : cameras,
	"reaper" : reaper
}

parsed = data

print("-----------------------------------")

default="./nvr-gen.json"

f = open(default, 'w')
f.write(json.dumps(parsed, indent=4, sort_keys=True))
f.close()

print("The file {} has been generated.  In order to use it, copy to /etc/nvr.json".format(default))
print("		sudo cp {} /etc/nvr.json".format(default))

init="./pynvr.gen"

f = open(init, 'w')
f.write( "#!/bin/bash\n")
f.write( "pynvr.py $1\n")
f.close()
print("The file {} has been generated.  In order to use it, run:".format(init))
print("		sudo cp {} /etc/init.d/pynvr".format(init))
print("		sudo chmod u+x /etc/init.d/pynvr".format(init))




"""Generic linux daemon base class for python 3.x."""

import sys, os, time, atexit, signal

class daemon:
	"""A generic daemon class.

	Usage: subclass the daemon class and override the run() method."""

	def __init__(self, pidfile): self.pidfile = pidfile
	
	def daemonize(self):
		"""Deamonize class. UNIX double fork mechanism."""

		try: 
			pid = os.fork() 
			if pid > 0:
				# exit first parent
				sys.exit(0) 
		except OSError as err: 
			sys.stderr.write('fork #1 failed: {0}\n'.format(err))
			sys.exit(1)
	
		# decouple from parent environment
		os.chdir('/') 
		os.setsid() 
		os.umask(0) 
	
		# do second fork
		try: 
			pid = os.fork() 
			if pid > 0:

				# exit from second parent
				sys.exit(0) 
		except OSError as err: 
			sys.stderr.write('fork #2 failed: {0}\n'.format(err))
			sys.exit(1) 
	
		# redirect standard file descriptors
		sys.stdout.flush()
		sys.stderr.flush()
		si = open(os.devnull, 'r')
		so = open(os.devnull, 'a+')
		se = open(os.devnull, 'a+')

		os.dup2(si.fileno(), sys.stdin.fileno())
		os.dup2(so.fileno(), sys.stdout.fileno())
		os.dup2(se.fileno(), sys.stderr.fileno())
	
		# write pidfile
		atexit.register(self.delpid)

		pid = str(os.getpid())
		with open(self.pidfile,'w+') as f:
			f.write(pid + '\n')
	
	def delpid(self):
		os.remove(self.pidfile)

	def start(self):
		"""Start the daemon."""

		# Check for a pidfile to see if the daemon already runs
		try:
			with open(self.pidfile,'r') as pf:

				pid = int(pf.read().strip())
		except IOError:
			pid = None
	
		if pid:
			message = "pidfile {0} already exist. " + \
					"Daemon already running?\n"
			sys.stderr.write(message.format(self.pidfile))
			sys.exit(1)
		
		# Start the daemon
		self.daemonize()
		self.run()

	def status(self):
		"""Status of the daemon."""

		# Get the pid from the pidfile
		try:
			with open(self.pidfile,'r') as pf:
				pid = int(pf.read().strip())
		except IOError:
			pid = None
	
		if not pid:
			sys.stderr.write("pidfile {0} does not exist. Daemon not running.\n".format(self.pidfile))
			return # not an error in a restart

		sys.stderr.write("Daemon running with pid {}\n".format(pid))

	def stop(self):
		"""Stop the daemon."""

		# Get the pid from the pidfile
		try:
			with open(self.pidfile,'r') as pf:
				pid = int(pf.read().strip())
		except IOError:
			pid = None
	
		if not pid:
			message = "pidfile {0} does not exist. " + \
					"Daemon not running?\n"
			sys.stderr.write(message.format(self.pidfile))
			return # not an error in a restart

		# Try killing the daemon process	
		try:
			while 1:
				os.kill(pid, signal.SIGTERM)
				time.sleep(0.1)
		except OSError as err:
			e = str(err.args)
			if e.find("No such process") > 0:
				if os.path.exists(self.pidfile):
					os.remove(self.pidfile)
			else:
				print (str(err.args))
				sys.exit(1)

	def restart(self):
		"""Restart the daemon."""
		self.stop()
		self.start()

	def run(self):
		"""You should override this method when you subclass Daemon.
		
		It will be called after the process has been daemonized by 
		start() or restart()."""




import subprocess
import os
import platform
import glob
import time

def getPercent(filename, logger):
	try:
		df = subprocess.Popen(["df",filename], stdout=subprocess.PIPE)
		output = df.communicate()[0].decode(encoding='UTF-8').split("\n")[1].split()

		percent = "0%"
		mountpoint = None
		if (platform.system() == "Darwin"):
			device, blocks, used, available, percent, iused, ifree, percentiused, mountpoint = output
		elif (platform.system() == "Linux"):
			device, size, used, available, percent, mountpoint = output
		else:
			logger.error("unsupported system")
	
		percent = percent.split("%")[0]
	
		return percent,mountpoint
	except:
		return 0, None
	
def validate(config, logger):
	m = {}
	for p in config:
		disk = p['disk']
		percent = p['percent']
		actual_percent, mountpoint = getPercent(disk, logger)
		if mountpoint != None:
			if mountpoint in m.keys():
				if (percent != m[mountpoint]):
					logger.error("disk {} with different percent({} vs {}) on same mountpoint {} as other reap locations.".format(disk, percent, m[mountpoint], mountpoint))
					return False
			m[mountpoint] = percent
	return True

def reap(config, logger):
	if config == None:
		return False

	workDone = False

	# For a given mount point, find all locations and files
	filesToReap = {}

	for p in config:
		location = p['disk']
		trigger_percent = p['percent'].split("#")[0]
		extensions = p['extensions']

		percent, mountpoint = getPercent(location, logger)
		if mountpoint is None:
			continue

		# Remove empty directories
		dirs = [d for d in glob.glob(location + "/????????") if os.path.isdir(d)]
		for dir in dirs:
			if not os.listdir(dir):
				logger.info("removing empty directory {}".format(dir))
				os.rmdir(dir)

		if (int(percent) > int(trigger_percent)):
			files = []
			for ext in extensions:
				files.extend(glob.glob(location + "/????????/*." + ext))


			if mountpoint in filesToReap.keys():
				filesToReap[mountpoint].extend(files)
			else:
				filesToReap[mountpoint] = files


	for mountpoint, files in filesToReap.items():
		for i in range(0,5):
			if len(files)>0:
				workDone = True
				file = min(files, key=os.path.getctime)
				files.remove(file)
	
				logger.info("removing {}".format(file))
				os.remove(file)

	return workDone




#!/usr/bin/env python3
import json
import pprint
from os.path import isdir,exists
from shared import reap

def validate_dir(json, key, logger, error):
	if (key not in json.keys()):
		logger.error("key {} not found in stanza {}".format( key, json))
		return False

	if (not isdir(json[key])):	
		if error:
			logger.error('{} does not exists: {}'.format(key, json[key]))
			return False
		else:
			logger.error('{} does not exists: {}'.format(key, json[key]))

	return True

def locate_config_file(_filename):
	if (exists(_filename)):
		filename = _filename
	else:
		filename = "/etc/" + _filename
	return filename


def parse(filename, logger):
	data = {}

	with open(filename) as data_file:    
    		data = json.load(data_file)

	# Validate
	if 'system' not in data.keys():
		logger.error("no system stanza.")
		return None
	if 'server' not in data['system'].keys():
		logger.error("no system.server stanza.")
		return None
	if 'cameras' not in data.keys():
		logger.error("no cameras stanza.")
		return None

	s = data['system']

	if not validate_dir(data['system'],'jpg_cache_dir', logger, True):
		return None
	if not validate_dir(data['system'],'mp4_cache_dir', logger, True):
		return None
	if not validate_dir(data['system'],'camera_log_dir', logger, True):
		return None
	if not validate_dir(data['system'],'log_dir', logger, True):
		return None

	for camera in data['cameras']:
		if camera['enabled']:
			if not validate_dir(camera, 'dst', logger, True):
				return None

	if 'reaper' in data.keys():
		if (not reap.validate(data['reaper'], logger)):
			return None	

	return data





import re
import os
import glob
import time
from datetime import datetime,timedelta
from os import listdir
from collections import namedtuple
from shared import common
from datetime import datetime,timedelta

DateRange = namedtuple("DateRange", "begin end")
Chunk = namedtuple("Chunk", "mp4 jpg thumbnail date")

def minutesBetween(daterange):
	dates = []

	diff = int((daterange.end - daterange.begin).total_seconds() / 60)
	for a in range(0,diff+1):
		dates.append(daterange.begin + timedelta(minutes = a))

	return dates

def rangesFromDates(dates):
	ranges = []

	if dates != None and len(dates) > 0:
		dates = set(dates)
		dates = list(dates)
		dates.sort()

		begin = min(dates)
		lastdate = begin
		for date in dates:
			diff = abs(lastdate - date)
			minutes = int(diff.total_seconds()/60)
			if minutes > 1:
				ranges.append(DateRange(begin, lastdate))
				begin = date
			lastdate = date

		ranges.append(DateRange(begin,lastdate))

	return ranges

def mergeRanges(ranges):

	dates = []

	for range in ranges:
		dates.extend(minutesBetween(range))

	return rangesFromDates(dates)




class Camera:
	def __init__(self, config):
		self.config = config
		self.enabled = config['enabled']
		self.url = config['url']
		self.path = config['dst']
		self.camera_id = config['id']

	def cameraMP4s(self):
		return glob.glob(os.path.join(self.path,"/*/{}.*.*.mp4".format(self.camera_id)))

	def cameraMP4sForHour(self,date):
		return glob.glob(os.path.join(common.get_dated_dir(self.path,date),"{}.{}??.mp4".format(self.camera_id, date.strftime("%Y%m%d.%H"))))

	def cameraMP4sForDay(self,date):
		return glob.glob(os.path.join(common.get_dated_dir(self.path,date),"{}.{}.????.mp4".format(self.camera_id, date.strftime("%Y%m%d"))))

	def getChunk(self, date):
		mp4 = common.getMP4Filename(self.config, date)
		#if not os.path.isfile(mp4):
			#mp4 = None

		jpg = common.getJPGFilename(common.config, self.config, date)
		#if not os.path.isfile(jpg):
			#jpg = None

		thumbnail = common.getThumbnailFilename(common.config, self.config, date)
		#if not os.path.isfile(thumbnail):
			#thumbnail = None

		if mp4 == None and jpg == None and thumbnail == None:
			return None

		return Chunk(mp4, jpg, thumbnail, date)

	def dateRanges(self):
		mp4s = self.cameraMP4s()
		dates = [datetime.strptime(os.path.basename(c), "{}.%Y%m%d.%H%M.mp4".format(self.camera_id)) for c in mp4s]

		return rangesFromDates(dates)

	def dateRangesForHour(self, date):
		mp4s = self.cameraMP4sForHour(date)
		dates = [datetime.strptime(os.path.basename(c), "{}.%Y%m%d.%H%M.mp4".format(self.camera_id)) for c in mp4s]
		return rangesFromDates(dates)

	def dateRangesForDay(self, date):
		mp4s = self.cameraMP4sForDay(date)
		dates = [datetime.strptime(os.path.basename(c), "{}.%Y%m%d.%H%M.mp4".format(self.camera_id)) for c in mp4s]
		return rangesFromDates(dates)

	def maskedURL(self):
		return re.sub(r'rtsp:\/\/.*@', 'rtsp://........@', self.url, flags=re.IGNORECASE)

	def URL(self):
		return self.url

	def isRecording(self):
		chunks = glob.glob(os.path.join(self.path,"*/{}.*.*.chunk".format(self.camera_id)))
		chunks = [a for a in chunks if os.stat(a).st_size != 0]
		if len(chunks) == 0:
			return False

		dates = [datetime.strptime(os.path.basename(c), "{}.%Y%m%d.%H%M.chunk".format(self.camera_id)) for c in chunks]
		dates.sort()
		end = max(dates)

		d1_ts = time.mktime(end.timetuple())
		d2_ts = time.mktime(datetime.now().timetuple())
		minutes = int(d2_ts-d1_ts) / 60

		return minutes < 2



import subprocess
import os
import platform

prereqs_msg = ""
ffmpeg_path = None

def subprocess_mp4_to_jpg(mp4, jpg, options):
	cmd = "{} -i {} -vframes 1 {} 2>/dev/null 1>/dev/null".format(ffmpeg_path, mp4, jpg)
	return subprocess.Popen(cmd, shell=True)

def subprocess_chunk_to_jpg(basepath, chunk, jpg, options):
	cmd = "{}/extract_h264.py {} | {} -f h264 -r 30 -i - -vcodec copy  -f mpegts - 2>/dev/null | {} -r 30 -i - -vframes 1 -f {} image2 {} 2>/dev/null 1>/dev/null".format(basepath, chunk, ffmpeg_path, ffmpeg_path, options, jpg)
	return subprocess.Popen(cmd, shell=True)

def subprocess_chunk_to_mp4(basepath, chunk, mp4, options):
	cmd = "{}/extract_h264.py {} | {} -f h264 -r 30 -i - -vcodec copy  -f mp4 {} -movflags +faststart 2>/dev/null".format(basepath, chunk, ffmpeg_path, mp4)
	return subprocess.Popen(cmd, shell=True)

def subprocess_resize_jpg(from_jpg, to_jpg, options):
	# scale=320:-1
	cmd = "{} -i {} -vf {} {} 2>/dev/null".format(ffmpeg_path, from_jpg, options, to_jpg)
	return subprocess.Popen(cmd, shell=True)

def subprocess_chunks_to_mp4(basepath, chunks, mp4):
	cmd = "{}/extract_h264.py {} | {} -f h264 -r 30 -i - -vcodec copy  -f mp4 -movflags +faststart {} 2>/dev/null".format(basepath, ' '.join(str(x) for x in chunks), ffmpeg_path, mp4)
	return subprocess.Popen(cmd, shell=True)

def _date_string(temp_date):
	if (type(temp_date) is str):
		return temp_date
	return temp_date.strftime("%Y%m%d.%H%M")

def _dir_string(temp_date):
	if (type(temp_date) is str):
		return temp_date.split(".")[0]
	return temp_date.strftime("%Y%m%d")

def get_dated_dir(temp_dir, temp_date):
	directory = os.path.join(temp_dir, _dir_string(temp_date))
	if not os.path.exists(directory):
		os.makedirs(directory)
	return directory
	

def getChunkFilename(camera, temp_date):
	return get_dated_dir(camera['dst'], temp_date) + "/{}.{}.chunk".format(camera['id'],_date_string(temp_date))
def getChunkFilenameBy(path, camera_id, temp_date):
	return get_dated_dir(path,temp_date) + "/{}.{}.chunk".format(camera_id,_date_string(temp_date))

def getJPGFilename(config, camera, temp_date):
	d = get_dated_dir(config['system']['jpg_cache_dir'],temp_date)
	return "{}/{}.preview.{}.jpg".format(d, camera['id'],_date_string(temp_date))
def getJPGFilenameBy(path, camera_id, temp_date):
	return "{}/{}.preview.{}.jpg".format(get_dated_dir(path,temp_date), camera_id,_date_string(temp_date))

def getThumbnailFilename(config, camera, temp_date):
	d = get_dated_dir(config['system']['jpg_cache_dir'],temp_date)
	return "{}/{}.thumbnail.{}.jpg".format(d, camera['id'],_date_string(temp_date))
def getThumbnailFilenameBy(path, camera_id, temp_date):
	return "{}/{}.thumbnail.{}.jpg".format(get_dated_dir(path, temp_date), camera_id,_date_string(temp_date))

def getMP4Filename(camera, temp_date):
	return "{}/{}.{}.mp4".format(get_dated_dir(camera['dst'],temp_date), camera['id'],_date_string(temp_date))
def getMP4FilenameBy(path, camera_id, temp_date):
	return "{}/{}.{}.mp4".format(get_dated_dir(path, temp_date), camera_id,_date_string(temp_date))

def getMP4FilenameRange(config, camera, temp_from_date, temp_to_date):
	d = config['system']['mp4_cache_dir']
	return "{}/{}.range.{}.to.{}.mp4".format(d, camera['id'],_date_string(temp_from_date), _date_string(temp_to_date))
def getMP4FilenameRangeBy(path, camera_id, temp_from_date, temp_to_date):
	return "{}/{}.range.{}.to.{}.mp4".format(path, camera['id'],_date_string(temp_from_date), _date_string(temp_to_date))

def getCameraConfig(camera_id):
	global config
	for cam in config['cameras']:
		if cam['id'] == camera_id:
			return cam
	return None

config = None

def setConfig(c):
	global config
	config = c

def getConfig():
	return config

def which(program):
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None

def check_prereqs():
	global ffmpeg_path, prereqs_msg
	ffmpeg1 = which("ffmpeg") 
	ffmpeg_path = ffmpeg1 if ffmpeg1 != None else which("ffmpeg.exe")

	if ffmpeg_path == None:
		prereqs_msg = prereqs_msg + "ffmpeg was not found on the path, perhaps it isn't installed? "

	return True if len(prereqs_msg) == 0 else False

def get_prereqs_msg():
	return prereqs_msg

def get_ffmpeg_path():
	return ffmpeg_path
	
	



from setuptools import setup, find_packages, Extension

import os

INCLUDE_DIR = '/usr/include'
INSTALL_DIR = './live'
dirs=['%s/%s' % (INCLUDE_DIR, x) for x in ['liveMedia', 'BasicUsageEnvironment', 'UsageEnvironment', 'groupsock']]
dirs.extend(['%s/%s/include' % (INSTALL_DIR, x) for x in ['liveMedia', 'BasicUsageEnvironment', 'UsageEnvironment', 'groupsock']])
module = Extension('live555',
                   include_dirs=dirs,
                   libraries=['liveMedia', 'groupsock', 'BasicUsageEnvironment', 'UsageEnvironment'],
                   #extra_compile_args = ['-fPIC'],
                   library_dirs=['%s/%s' % (INSTALL_DIR, x) for x in ['liveMedia', 'UsageEnvironment','BasicUsageEnvironment', 'groupsock']],
                   sources = ['pylive555/module.cpp'])

setup(
	name = "pynvr",
	version = "0.1.8",
	description = "Network Video Recorder for recording h264 rtsp IP cameras",
	author = "Jack Strohm",
	author_email = "hoyle.hoyle@gmail.com",
	url = "https://gitlab.com/hoyle.hoyle/pynvr",
	license = "MIT",
	packages = find_packages(),
	scripts = ['pynvr.py', 'rtsp.py', 'extract_h264.py', 'pynvr_config.py'],
	install_requires = ['bottle>=0.12'],
	include_package_data = True,
	package_data = {
		'': ['*.txt', '*.rst', '*.md'],
		'static': ['*.html','*.png', '*.css', '*.gif', '*.html', '*.js', '*.png' ],
	},
	ext_modules = [module],
	keywords = ['nvr', 'security', 'recording', 'ip camera', 'rtsp', 'h264', 'camera'],
)

